! ***********************************************************************
!
!   Copyright (C) 2010 Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
module run_star_extras

  use star_lib
  use star_def
  use const_def

  use gyre_support

  implicit none

  integer, save  :: i_mode_lo
  integer, save  :: i_mode_hi

  real(dp), save       :: growth_thresh
  character(256), save :: strip_file
  integer, save        :: strip_unit

  real(dp), save :: Teff_thresh
  real(dp), save :: Teff_tol

  real(dp), allocatable, save :: freqs(:)
  real(dp), allocatable, save :: growths(:)

contains
  
  subroutine extras_controls (id, ierr)

    integer, intent(in)  :: id
    integer, intent(out) :: ierr

    type (star_info), pointer :: s

    ierr = 0

    call star_ptr(id, s, ierr)
    if (ierr /= 0) return
         
    ! this is the place to set any procedure pointers you want to change
    ! e.g., other_wind, other_mixing, other_energy  (see star_data.inc)
         
  end subroutine extras_controls

!****
      
  integer function extras_startup (id, restart, ierr)

    integer, intent(in)  :: id
    logical, intent(in)  :: restart
    integer, intent(out) :: ierr

    type (star_info), pointer :: s

    ierr = 0

    call star_ptr(id, s, ierr)
    if (ierr /= 0) return

    extras_startup = 0

    if (.not. restart) then
       call alloc_extra_info(s)
    else ! it is a restart
       call unpack_extra_info(s)
    end if

    ! Initialize GYRE

    call init_gyre(ierr)

    ! If necessary, open the strip output file

    if (strip_file /= '') then
       open(NEWUNIT=strip_unit, FILE=strip_file, STATUS='REPLACE')
    endif

  end function extras_startup
      
!****

  integer function extras_check_model (id, id_extra) result (action)
 
    integer, intent(in) :: id, id_extra

    action = extras_check_model_multi(id, id_extra)

    return

  end function extras_check_model

!****

  integer function extras_check_model_multi (id, id_extra) result (action)

    integer, intent(in) :: id, id_extra
    integer             :: ierr

    type(star_info), pointer   :: s
    logical, save              :: past_thresh = .FALSE.
    integer                    :: ipar(1)
    real(WP)                   :: rpar(1)
    character(64), save        :: state = 'START'
    logical, allocatable, save :: unstab_mask(:)
    logical, allocatable, save :: unstab_mask_prev(:)
    real(WP), save             :: delta_t
    real(WP), save             :: Teff_a
    real(WP), save             :: Teff_b
    real(WP), save             :: x_a
    real(WP), save             :: x_b
    real(WP), save             :: x_new
    real(WP), save             :: x_root
    logical, allocatable, save :: unstab_mask_a(:)
    logical, allocatable, save :: unstab_mask_b(:)

    ierr = 0

    call star_ptr(id, s, ierr)
    if (ierr /= 0) return

    action = keep_going

    ! Decide whether we're past Teff_thresh

    if (.NOT. past_thresh) then
       if (s%Teff < Teff_thresh) then
          past_thresh = .TRUE.
       else
          return
       endif
    endif

    ! Store the model for GYRE

    call store_model_for_gyre(s, keep_surface_point, add_atmosphere, ierr)
    
    if (ierr /= 0) then
       write(*,*) 'failed in store_model_for_gyre'
       return
    end if

    ! Run GYRE

    print *,'Running gyre...'

    if (allocated(freqs)) deallocate(freqs)
    if (allocated(growths)) deallocate(growths)

    allocate(freqs(0))
    allocate(growths(0))

    call gyre_get_modes(0, 'gyre.in', .TRUE., call_back, ipar, rpar)

    unstab_mask = growths > growth_thresh

    print *,'   ...got mask:', unstab_mask(i_mode_lo:i_mode_hi)
      
    if (i_mode_hi > SIZE(unstab_mask)) stop 'i_mode_hi exceeds number of modes'

    ! (Possibly) transition between states

    select case (state)

    case ('START')
       
       state = 'EVOLVE'

       unstab_mask_prev = unstab_mask

    case ('EVOLVE')

       ! Check for a change in the mode state

       if (ANY(unstab_mask(i_mode_lo:i_mode_hi) .NEQV. unstab_mask_prev(i_mode_lo:i_mode_hi))) then

          delta_t = s%dt

          ! Set up the initial bracket

          Teff_a = s%Teff_old
          Teff_b = s%Teff
          
          x_a = 0._WP
          x_b = 1._WP

          unstab_mask_a = unstab_mask_prev
          unstab_mask_b = unstab_mask

          state = 'BRACK'

       else

          unstab_mask_prev = unstab_mask

       endif

    case ('BRACK')

       ! Update the bracket

       if (ALL(unstab_mask(i_mode_lo:i_mode_hi) .EQV. &
               unstab_mask_b(i_mode_lo:i_mode_hi))) then

          Teff_b = s%Teff

          x_b = x_new

          unstab_mask_b = unstab_mask

       else

          Teff_a = s%Teff

          x_a = x_new

          unstab_mask_a = unstab_mask

       endif

       ! Check for convergence

       if (ABS(Teff_a - Teff_b) < Teff_tol) then

          x_root = x_new

          print *,'Found root at:', x_root

          if (strip_file /= '') then
             write(strip_unit, 100) s%model_number, s%Teff, s%L_phot, &
                  unstab_mask_a(i_mode_lo:i_mode_hi), unstab_mask_b(i_mode_lo:i_mode_hi)
100          format(I0,1X,E16.8,1X,E16.8,999(1X,1L))
          endif

          s%dt = delta_t*(1._WP - x_root)
          action = keep_going

          state = 'START'

       endif

    end select

    ! Prepare for the next step

    if (state == 'BRACK') then

       ! Calculate a new timestep using bisection

       x_new = 0.5_WP*(x_a + x_b)

       s%dt = delta_t*x_new
       action = redo

    endif

    ! Finish

    return

  end function extras_check_model_multi

!****

  subroutine call_back (md, ipar, rpar, ierr)
    
    type(mode_t), intent(in) :: md
    integer, intent(inout)   :: ipar(:)
    real(dp), intent(inout)  :: rpar(:)
    integer, intent(out)     :: ierr

    complex(dp) :: freq

    ! Store the mode frequency and growth rate (modes are found in
    ! order of increasing freq)

    freq = md%freq('PER_DAY')

    freqs = [freqs,REAL(freq)]
    growths = [growths,-2.*PI*AIMAG(freq)]

    ! Finish

    return

  end subroutine call_back

!****

  integer function how_many_extra_history_columns(id, id_extra)

    integer, intent(in) :: id, id_extra

    how_many_extra_history_columns = 0

  end function how_many_extra_history_columns

!****    
      
  subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)

    integer, intent(in) :: id, id_extra, n
    character (len=maxlen_history_column_name) :: names(n)
    real(dp) :: vals(n)
    integer, intent(out) :: ierr

    ierr = 0

  end subroutine data_for_extra_history_columns

!****

  integer function how_many_extra_profile_columns(id, id_extra)

    use star_def, only: star_info
    integer, intent(in) :: id, id_extra
    integer :: ierr
    type (star_info), pointer :: s

    how_many_extra_profile_columns = 0

  end function how_many_extra_profile_columns

!****      
      
  subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)

    use star_def, only: star_info, maxlen_profile_column_name
    use const_def, only: dp

    integer, intent(in) :: id, id_extra, n, nz
    character (len=maxlen_profile_column_name) :: names(n)
    real(dp) :: vals(nz,n)
    integer, intent(out) :: ierr

    ierr = 0

  end subroutine data_for_extra_profile_columns

!****
  
  integer function extras_finish_step (id, id_extra)

    integer, intent(in) :: id, id_extra

    integer                   :: ierr
    type (star_info), pointer :: s

    ierr = 0

    call star_ptr(id, s, ierr)
    if (ierr /= 0) return

    extras_finish_step = keep_going

    call store_extra_info(s)

  end function extras_finish_step
      
!****
  
  subroutine extras_after_evolve (id, id_extra, ierr)

    integer, intent(in)  :: id, id_extra
    integer, intent(out) :: ierr

    type (star_info), pointer :: s

    ierr = 0

    call star_ptr(id, s, ierr)
    if (ierr /= 0) return

  end subroutine extras_after_evolve
      
!****      
  
  subroutine alloc_extra_info(s)

    type (star_info), pointer :: s

    integer, parameter :: extra_info_alloc = 1

    call move_extra_info(s,extra_info_alloc)

  end subroutine alloc_extra_info
      
!****      

  subroutine unpack_extra_info(s)

    type (star_info), pointer :: s

    integer, parameter :: extra_info_get = 2

    call move_extra_info(s,extra_info_get)

  end subroutine unpack_extra_info
      
!****
  
  subroutine store_extra_info(s)

    type (star_info), pointer :: s

    integer, parameter :: extra_info_put = 3

    call move_extra_info(s,extra_info_put)

  end subroutine store_extra_info
      
!****    

  subroutine move_extra_info(s,op)

    type (star_info), pointer :: s
    integer, intent(in)       :: op
         
    integer, parameter :: extra_info_alloc = 1
    integer, parameter :: extra_info_get = 2
    integer, parameter :: extra_info_put = 3

    integer :: i, j, num_ints, num_dbls, ierr
         
    i = 0
    ! call move_int or move_flg    
    num_ints = i
         
    i = 0
    ! call move_dbl       
    
    num_dbls = i
         
    if (op /= extra_info_alloc) return
    if (num_ints == 0 .and. num_dbls == 0) return
    
    ierr = 0
    call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
    if (ierr /= 0) then
       write(*,*) 'failed in star_alloc_extras'
       write(*,*) 'alloc_extras num_ints', num_ints
       write(*,*) 'alloc_extras num_dbls', num_dbls
       stop 1
    end if
         
  contains
         
    subroutine move_dbl (dbl)

      real(dp) :: dbl

      i = i+1
      select case (op)
      case (extra_info_get)
         dbl = s% extra_work(i)
      case (extra_info_put)
         s% extra_work(i) = dbl
      end select

    end subroutine move_dbl
         
    subroutine move_int(int)

      integer :: int

      i = i+1
      select case (op)
      case (extra_info_get)
         int = s% extra_iwork(i)
      case (extra_info_put)
         s% extra_iwork(i) = int
      end select

    end subroutine move_int
         
    subroutine move_flg(flg)

      logical :: flg

      i = i+1
      select case (op)
      case (extra_info_get)
         flg = (s% extra_iwork(i) /= 0)
      case (extra_info_put)
         if (flg) then
            s% extra_iwork(i) = 1
         else
            s% extra_iwork(i) = 0
         end if
      end select
    end subroutine move_flg
      
  end subroutine move_extra_info

end module run_star_extras
